//---------------------------------------------------------------------------

#include <fstream.h>
#include <vcl.h>
#include <math.h>
#pragma hdrstop

#include "Main.h"
#include "NewImage.h"
#include "Scale.h"
#include "Rotate.h"
#include "Bright.h"
#include "Filter.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner)
{
        FileName = "";
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N4Click(TObject *Sender)
{
        Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N2Click(TObject *Sender)
{
        if (!SaveModified()) return;
        if (!OpenDialog->Execute()) return;
        Open(OpenDialog->FileName);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateImage()
{
        Graphics::TBitmap *bm = new Graphics::TBitmap();
        bm->Width = bi.biWidth;
        bm->Height = bi.biHeight;
        Image->Picture->Graphic = bm;
        HDC imgDC = Image->Canvas->Handle;
        HDC memDC = CreateCompatibleDC(imgDC);
        HBITMAP bmp = CreateDIBitmap(imgDC, &bi, CBM_INIT, buf, (BITMAPINFO *)&bi, DIB_RGB_COLORS);
        SelectObject(memDC, bmp);
        BitBlt(imgDC, 0, 0, bi.biWidth, bi.biHeight, memDC, 0, 0, SRCCOPY);
        DeleteDC(memDC);
        DeleteObject(bmp);
        delete bm;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateStatus()
{
        StatusBar->Panels->Items[0]->Text = AnsiString(": ") + AnsiString(bi.biWidth) + ", " +
                                AnsiString(": ") + AnsiString(bi.biHeight);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateTitle()
{
        Caption = FileName + " - " + " ";
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N5Click(TObject *Sender)
{
        if (FirstCreated) N6Click(NULL);
        else Save(FileName);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N6Click(TObject *Sender)
{
        SaveDialog->FileName = FileName;
        if (!SaveDialog->Execute()) return;
        Save(SaveDialog->FileName);
        FirstCreated = false;
}
//---------------------------------------------------------------------------
int __fastcall TMainForm::AdjustBufSize(int W, int H, int& rem)
{
        int s = W * 3;
        rem = (4 - (s & 3)) & 3;
        if (rem) s += rem;
        return s * H;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Open(AnsiString FileName)
{
        BITMAPFILEHEADER bf;
        BITMAPINFOHEADER bi;
        ifstream in;
        try
        {
                in.open(FileName.c_str(), ios::binary);
                if (!in) throw "  ";
                in.read((char *)&bf, 14/*sizeof(bf)*/);
                in.read((char *)&bi, sizeof(bi));
                if (bf.bfType != 0x4D42 || bi.biPlanes != 1)
                        throw "  ";
                if (bi.biCompression != BI_RGB || bi.biBitCount != 24)
                        throw "   ";
                if (bi.biWidth > 2048 || bi.biHeight > 2048)
                        throw "     2048 x 2048";
        }
        catch (char *msg)
        {
                if (in) in.close();
                Application->MessageBox(msg, "", MB_OK | MB_ICONSTOP);
                return;
        }
        bi.biSize = sizeof(bi);
        bf.bfOffBits = sizeof(bf) + bi.biSize;
        this->FileName = FileName;
        this->bf = bf;
        this->bi = bi;
        buf_size = AdjustBufSize(bi.biWidth, bi.biHeight, rem);
        if (buf) delete buf;
        buf = new char[buf_size];
//        in.seekg(bf.bfOffBits - 2, ios::beg);
        in.read(buf, buf_size);
        in.close();
        UpdateImage();
        UpdateStatus();
        UpdateTitle();
        Modified = false;
        FirstCreated = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Save(AnsiString FileName)
{
        ofstream out(FileName.c_str(), ios::binary);
        if (!out)
        {
                Application->MessageBox("  ", "", MB_OK | MB_ICONSTOP);
                return;
        }
        out.write((char *)&bf, sizeof(bf));
        out.write((char *)&bi, sizeof(bi));
        out.write(buf, buf_size);
        out.close();
        Modified = false;
        FirstCreated = false;
        this->FileName = FileName;
        UpdateTitle();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N7Click(TObject *Sender)
{
        if (!SaveModified()) return;
        if (NewImgDlg->ShowModal() != mrOk) return;
        bgcolor = NewImgDlg->color;
        NewImage(NewImgDlg->W, NewImgDlg->H, bgcolor);
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::CheckWH(AnsiString w, AnsiString h, int& W, int& H)
{
        try
        {
                W = StrToInt(w);
                H = StrToInt(h);
        }
        catch (EConvertError &e)
        {
                Application->MessageBox("       ", "", MB_OK | MB_ICONSTOP);
                return false;
        }
        if (W < 1 || W > 2048)
        {
                Application->MessageBox("       1  2048", "", MB_OK | MB_ICONEXCLAMATION);
                return false;
        }
        if (H < 1 || H > 2048)
        {
                Application->MessageBox("       1  2048", "", MB_OK | MB_ICONEXCLAMATION);
                return false;
        }
        return true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::NewImage(int W, int H, TColor c)
{
        if (buf) delete buf;
        buf = CreateTempImage(W, H, rem);
        Clear(c);
        FileName = "";
        UpdateImage();
        UpdateStatus();
        UpdateTitle();
        Modified = false;
        FirstCreated = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
        NewImage(256, 256, clWhite);
        bgcolor = clWhite;
}
//---------------------------------------------------------------------------
char* __fastcall TMainForm::CreateTempImage(int W, int H, int& rem)
{
        char *buf;
        buf_size = AdjustBufSize(W, H, rem);
        bf.bfType = 0x4D42;
        bf.bfSize = sizeof(bf) + sizeof(bi) + buf_size;
        bf.bfReserved1 = 0;
        bf.bfReserved2 = 0;
        bf.bfOffBits = bf.bfSize - buf_size;
        bi.biSize = sizeof(bi);
        bi.biWidth = W;
        bi.biHeight = H;
        bi.biPlanes = 1;
        bi.biBitCount = 24;
        bi.biCompression = BI_RGB;
        bi.biSizeImage = (W * 3 + rem) * H;
        bi.biXPelsPerMeter = 3780;
        bi.biYPelsPerMeter = 3780;
        bi.biClrUsed = 0;
        bi.biClrImportant = 0;
        buf = new char[buf_size];
        return buf;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N9Click(TObject *Sender)
{
        Invert();
        UpdateImage();
        Modified = true;
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::SaveModified()
{
        if (Modified)
        switch (Application->MessageBox(AnsiString("    \"" + FileName + "\"?").c_str(), "", MB_YESNOCANCEL | MB_ICONQUESTION))
        {
                case IDYES: N5Click(NULL); break;
                case IDCANCEL: return false;
        }
        return true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCloseQuery(TObject *Sender, bool &CanClose)
{
        CanClose = SaveModified();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N10Click(TObject *Sender)
{
        ColorDialog->Color = bgcolor;
        if (!ColorDialog->Execute()) return;
        Clear(ColorDialog->Color);
        UpdateImage();
        Modified = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Invert()
{
        int w = bi.biWidth, h = bi.biHeight, r = rem;
        char *b = buf;
        asm
        {
                cld
                mov esi, b
                mov ecx, h
@@5:            push ecx
                mov ecx, w
                mov edi, esi
@@6:            lodsw
                not ax
                stosw
                lodsb
                not al
                stosb
                loop @@6
                pop ecx
                add esi, r
                loop @@5
        }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Clear(TColor c)
{
        int w = bi.biWidth, h = bi.biHeight, r = rem;
        char *b = buf;
        asm
        {
                cld
                mov edi, b
                mov ecx, h
                mov bx, word ptr c
                xchg bl, bh
                mov dl, byte ptr c+2
@@3:            push ecx
                mov ecx, w
@@4:            mov al, dl
                stosb
                mov ax, bx
                stosw
                loop @@4
                pop ecx
                add edi, r
                loop @@3
        }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N11Click(TObject *Sender)
{
        ScaleDlg->WidthEdit->Text = bi.biWidth;
        ScaleDlg->HeightEdit->Text = bi.biHeight;
        if (ScaleDlg->ShowModal() != mrOk) return;
        Scale(ScaleDlg->W, ScaleDlg->H);
        UpdateImage();
        UpdateStatus();
        Modified = true;
}
//---------------------------------------------------------------------------
double round(double x)
{
        if (x - floor(x) < 0.5) return floor(x);
        return ceil(x);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Rotate(float a, bool r)
{
        TColor color = bgcolor;
        int rem2, i, j;
        int w = bi.biWidth, W = w;
        int h = bi.biHeight, H = h;
        float c_ = cos(a * M_PI / 180);
        float s_ = sin(a * M_PI / 180);
        int c = int(c_ * 65536);
        int s = int(s_ * 65536);
        if (r)
        {
                W = int(round(fabs(c_) * w + fabs(s_) * h));
                H = int(round(fabs(s_) * w + fabs(c_) * h));
                if (W > 2048 ||H > 2048)
                {
                        Application->MessageBox("  ,       2048 x 2048", "", MB_OK | MB_ICONSTOP);
                        return;
                }
        }
        int W1 = W >> 1;
        int H1 = H >> 1;
        char *buf2 = CreateTempImage(W, H, rem2);
        char *b = buf;
        int scan1 = w * 3 + rem;
        int x, y, j1, j2, i1, i2;
        i1 = (w << 15) - W1 * c + H1 * s - (s + 1) * 0.5;
        i2 = -H1 * c + (h << 15) - W1 * s - (1 - s) * 0.5;
        asm
        {
                cld
                mov esi, b
                mov edi, buf2
                mov ecx, H
@@9:            push ecx
                mov ecx, W
                mov ebx, i1
                mov edx, i2
@@10:           push ebx
                push edx
                shr ebx, 16
                shr edx, 16
                test ebx, 80000000h
                jnz short @@11
                cmp ebx, w
                jnb short @@11
                test edx, 80000000h
                jnz short @@11
                cmp edx, h
                jnb short @@11
                mov eax, ebx
                shl ebx, 1
                add ebx, eax
                imul edx, scan1
                add ebx, edx
                mov ax, [esi + ebx]
                stosw
                mov al, [esi + ebx + 2]
                stosb
                jmp short @@12
@@11:           mov al, byte ptr color+2
                stosb
                mov ax, word ptr color
                xchg ah, al
                stosw
@@12:           pop edx
                pop ebx
                add ebx, c
                add edx, s
                loop @@10
                add edi, rem2
                mov eax, s
                sub i1, eax
                mov eax, c
                add i2, eax
                pop ecx
                loop @@9
        }
        rem = rem2;
        delete buf;
        buf = buf2;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Scale(int W, int H)
{
        int rem2;
        int w = bi.biWidth;
        int h = bi.biHeight;
        int scan = w * 3 + rem;
        char *buf2 = CreateTempImage(W, H, rem2);
        char *b1 = buf;
        int k1 = (w << 16) / W;
        int k2 = (h << 16) / H;
        asm
        {
                cld
                mov edi, buf2
                mov ecx, H
                xor edx, edx
@@7:            push ecx
                mov esi, edx
                shr esi, 16
                imul esi, scan
                add esi, b1
                xor ebx, ebx
                mov ecx, W
@@8:            push ebx
                push esi
                shr ebx, 16
                mov eax, ebx
                shl eax, 1
                add ebx, eax
                add esi, ebx
                movsw
                movsb
                pop esi
                pop ebx
                add ebx, k1
                loop @@8
                pop ecx
                add edi, rem2
                add edx, k2
                loop @@7
        }
        rem = rem2;
        delete buf;
        buf = buf2;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FlipH()
{
        int w1 = bi.biWidth >> 1, h = bi.biHeight;
        int scan = bi.biWidth * 3 + rem;
        int k = scan - rem - 3;
        char *b = buf;
        asm
        {
                cld
                mov esi, b
                mov ecx, h
@@1:            push ecx
                push esi
                mov edi, esi
                add edi, k
                mov ecx, w1
@@2:            mov ax, [esi]
                xchg ax, [edi]
                mov [esi], ax
                mov al, [esi+2]
                xchg al, [edi+2]
                mov [esi+2], al
                add esi, 3
                sub edi, 3
                loop @@2
                pop esi
                add esi, scan
                pop ecx
                loop @@1
        }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FlipV()
{
        int w = bi.biWidth, h1 = bi.biHeight >> 1;
        int r = rem;
        int scan = w * 3 + r;
        int k = buf_size - scan;
        char *b = buf;
        asm
        {
                cld
                mov esi, b
                mov edi, esi
                add edi, k
                mov ecx, h1
@@13:           push ecx
                push edi
                mov ecx, w
@@14:           mov ax, [esi]
                xchg [edi], ax
                mov [esi], ax
                inc esi
                inc esi
                inc edi
                inc edi
                mov al, [esi]
                xchg [edi], al
                mov [esi], al
                inc esi
                inc edi
                loop @@14
                pop edi
                add esi, r
                sub edi, scan
                pop ecx
                loop @@13
        }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N12Click(TObject *Sender)
{
        FlipH();
        UpdateImage();
        Modified = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N14Click(TObject *Sender)
{
        if (RotateDlg->ShowModal() != mrOk) return;
        Rotate(RotateDlg->a, RotateDlg->resize);
        UpdateImage();
        UpdateStatus();
        Modified = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N13Click(TObject *Sender)
{
        FlipV();
        UpdateImage();
        Modified = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N16Click(TObject *Sender)
{
        if (BrightDlg->ShowModal() != mrOk) return;
        Brightness();
        UpdateImage();
        Modified = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Brightness()
{
        unsigned char *trans_table = new unsigned char[768];
        char *b = buf;
        int c;
        int w = bi.biWidth, h = bi.biHeight;
        int r = rem;
        for (int i = 0; i < 3; i++)
                for (int j = 0; j < 256; j++)
                {
                        c = (j - 128) * BrightDlg->channels[i].k +
                                BrightDlg->channels[i].b + 128;
                        if (c < 0) c = 0;
                        if (c > 255) c = 255;
                        trans_table[i * 256 + j] = c;
                }
        asm
        {
                cld
                mov esi, b
                mov ebx, trans_table
                xor eax, eax
                mov ecx, h
@@15:           push ecx
                mov edi, esi
                mov ecx, w
@@16:           lodsb
                mov al, [ebx + eax + 512]
                stosb
                lodsb
                mov al, [ebx + eax + 256]
                stosb
                lodsb
                mov al, [ebx + eax]
                stosb
                loop @@16
                add esi, r
                pop ecx
                loop @@15
        }
        delete trans_table;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N18Click(TObject *Sender)
{
        Blur();
        UpdateImage();
        Modified = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Blur()
{
        int w = bi.biWidth - 1, h = bi.biHeight - 1;
        if (w < 1 || h < 1) return;
        int r = rem;
        int scan = w * 3 + r;
        char *b = buf;
        asm
        {
                cld
                mov ebx, scan
                xor edx, edx
                mov esi, b
                mov ecx, h
@@17:           push ecx
                mov edi, esi
                mov ecx, w
@@18:           xor eax, eax
                xor edx, edx
                lodsb
                mov dl, [esi + 2]
                add eax, edx
                mov dl, [esi + ebx - 1]
                add eax, edx
                mov dl, [esi + ebx + 2]
                add eax, edx
                shr eax, 2
                stosb
                xor eax, eax
                xor edx, edx
                lodsb
                mov dl, [esi + 2]
                add eax, edx
                mov dl, [esi + ebx - 1]
                add eax, edx
                mov dl, [esi + ebx + 2]
                add eax, edx
                shr eax, 2
                stosb
                xor eax, eax
                xor edx, edx
                lodsb
                mov dl, [esi + 2]
                add eax, edx
                mov dl, [esi + ebx - 1]
                add eax, edx
                mov dl, [esi + ebx + 2]
                add eax, edx
                shr eax, 2
                stosb
                loop @@18
                add esi, r
                pop ecx
                loop @@17
        }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::N19Click(TObject *Sender)
{
        if (FilterDlg->ShowModal() != mrOk) return;
        UserFilter();
        UpdateImage();
        Modified = true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UserFilter()
{
        int w = bi.biWidth, h = bi.biHeight;
        int i, j, k, l, scan = w * 3 + rem, m;
        int i1, j1, i2, j2;
        int b1, e1, b2, e2;
        int s = FilterDlg->Size, s2 = s / 2;
        int k1 = s2 * scan, k2 = s2 * 3, k3;
        float sum[3], C[7][7];
        char *buf2 = CreateTempImage(w, h, rem);
        memcpy(C, FilterDlg->C, sizeof(C));
        memcpy(buf2, buf, buf_size);
        if (w < s || h < 3) return;
        i1 = s2 * scan;
        for (i = s2; i < h - s2; i++)
        {
                j1 = s2 * 3;
                for (j = s2; j < w - s2; j++)
                {
                        k3 = j1 - k2;
                        for (m = 0; m < 3; m++, k3++)
                        {
                                sum[m] = 0;
                                i2 = i1 - k1;
                                for (k = 0; k < s; k++)
                                {
                                        j2 = i2 + k3;
                                        for (l = 0; l < s; l++)
                                        {
                                                sum[m] += C[l][k] * (unsigned char)buf[j2];
                                                j2 += 3;
                                        }
                                        i2 += scan;
                                }
                                if (sum[m] < 0) sum[m] = 0;
                                if (sum[m] > 255) sum[m] = 255;
                                buf2[i1 + j1 + m] = sum[m];
                        }
                        j1 += 3;
                }
                i1 += scan;
        }
        delete buf;
        buf = buf2;
}
//---------------------------------------------------------------------------

